use crate::{Card, error};
use crate::iso7816_4::apducommand::{APDUCommand, IsoCase};
use crate::desfire::apduinstructions::APDUInstructions;
use crate::crypto::cipher::tdes::Tdes;
use crate::crypto::cipher::aes::AES;
use crate::crypto::cipher::Cipher;
use crate::crypto::util;
use crate::error::{Result, Error};
use crate::error::Error::{InvalidApplicationID, InvalidKeyID, InvalidFileID, NumKeys};
use crate::crypto::util::expand_to_blocksize;
use crate::desfire::{FileCommunication, ChangeMasterKeySettings, CreateDeleteFile, FileDirectoryAccess, ChangeMasterKey, ChangeApplicationKey, CryptoOperationsType, FileIdentifiers};
use crate::iso7816_4::apduresponse::APDUResponse;
use num_traits::FromPrimitive;
use std::fs::read_to_string;

pub struct Desfire {
    card: Box<dyn Card>,
    session_key: Option<Vec<u8>>,
    cbc_iv: Option<Vec<u8>>,
}

impl Desfire {
    // Desfire commands

    /// <summary>
    /// Select Application by ApplicationID (AID)
    /// </summary>
    /// <param name="aid">3 Byte AID</param>
    pub fn select_application(&self, aid: u32) -> Result<()> {
        if aid > 0xFFFFFF {
            return Err(InvalidApplicationID);
        }

        let cmd_select_app = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::SELECT_APPLICATION as u8,
            data: Option::from(aid.to_le_bytes()[..3].to_vec()), //FIXME: Which byteorder?
            ..Default::default()
        };

        let response = self.card.transmit(cmd_select_app).unwrap();

        response.check().map_err(|e| e)
    }

    /// <summary>
    /// Authenticate to PICC, with ISO Authenticate for DES Key
    /// </summary>
    /// <param name="key_id">0x01 - 0x0D</param>
    /// <param name="key">Array of 8/16 Bytes</param>
    /// <param name="rndA">!!! WARNING For Testing only !!!</param>
    pub fn authenticate_iso_des(&mut self, key_id: u8, key: &[u8], rnd_a: Option<[u8; 8]>) -> Result<()> {
        if key_id > 0x0E {
            return Err(InvalidKeyID);
        }

        let cmd_challenge_request = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::AUTHENTICATE_ISO as u8,
            data: Option::from(vec!(key_id)),
            ..Default::default()
        };

        let response = self.card.transmit(cmd_challenge_request).unwrap();

        match response.check() {
            Ok(_) => {}
            Err(e) => { return Err(e); }
        }

        let rnd_b_enc = response.body.unwrap().as_slice();

        let mut rnd_b = Tdes::decrypt(rnd_b_enc, key, vec![0 as u8; 8].as_slice()).unwrap();

        rnd_b.rotate_left(1);

        let rnd_b_rl = rnd_b.as_slice();

        let rnd_a = match rnd_a {
            None => { rand::random() }
            Some(i) => { i }
        };

        let rnd_ab = [&rnd_a, rnd_b_rl].concat();

        let rnd_ab_enc = Tdes::encrypt(rnd_ab.as_slice(), key, rnd_b_enc).unwrap();

        let cmd_challenge_response = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::CONTINUE as u8,
            data: Some(rnd_ab_enc),
            ..Default::default()
        };

        let response = self.card.transmit(cmd_challenge_response).unwrap();

        match response.check() {
            Ok(_) => {}
            Err(e) => { return Err(e); }
        }

        let mut rnd_a_rot_from_card = Tdes::decrypt(response.body.unwrap().as_slice(), key, rnd_b.as_slice()).unwrap();
        rnd_a_rot_from_card.rotate_right(1);
        let rnd_a_from_card = rnd_a_rot_from_card.as_slice();

        if rnd_a != rnd_a_from_card {
            return Err(Error::InvalidPICCChallenge);
        }

        self.session_key = Some(generate_session_key_des(&rnd_a, rnd_b.as_slice()).unwrap());

        self.cbc_iv = Some(vec![0 as u8; 8]);
        Ok(())
    }

    /// <summary>
    /// Authenticate to PICC, with ISO Authenticate for AES Key
    /// </summary>
    /// <param name="key_id">0x01 - 0x0D</param>
    /// <param name="key">Array of 8/16 Bytes</param>
    /// <param name="rndA">!!! WARNING For Testing only !!!</param>
    pub fn authenticate_iso_aes(&mut self, key_id: u8, key: &[u8], rnd_a: Option<[u8; 8]>) -> Result<()> {
        if key_id > 0x0E {
            return Err(InvalidKeyID);
        }

        let auth_iv: Vec<u8>;

        let cmd_challenge_request = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::AUTHENTICATE_AES as u8,
            data: Option::from(vec!(key_id)),
            ..Default::default()
        };

        let response = self.card.transmit(cmd_challenge_request).unwrap();

        match response.check() {
            Ok(_) => {}
            Err(e) => { return Err(e); }
        }

        let rnd_b_enc = response.body.unwrap().as_slice();

        let mut rnd_b = AES::decrypt(rnd_b_enc, key, vec![0 as u8; 16].as_slice()).unwrap();

        auth_iv = rnd_b.clone();

        rnd_b.rotate_left(1);

        let rnd_b_rl = rnd_b.as_slice();

        let rnd_a = match rnd_a {
            None => { rand::random() }
            Some(i) => { i }
        };

        let rnd_ab = [&rnd_a, rnd_b_rl].concat();

        let rnd_ab_enc = AES::encrypt(rnd_ab.as_slice(), key, auth_iv.as_slice()).unwrap();

        let cmd_challenge_response = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::CONTINUE as u8,
            data: Some(rnd_ab_enc),
            ..Default::default()
        };

        let response = self.card.transmit(cmd_challenge_response).unwrap();

        match response.check() {
            Ok(_) => {}
            Err(e) => { return Err(e); }
        }

        let mut rnd_a_rot_from_card = AES::decrypt(response.body.unwrap().as_slice(), key, auth_iv.as_slice()).unwrap();
        rnd_a_rot_from_card.rotate_right(1);
        let rnd_a_from_card = rnd_a_rot_from_card.as_slice();

        if rnd_a != rnd_a_from_card {
            return Err(Error::InvalidPICCChallenge);
        }

        self.session_key = Some(generate_session_key_aes(&rnd_a, rnd_b.as_slice()).unwrap());

        self.cbc_iv = Some(vec![0 as u8; 16]);
        Ok(())
    }

    /// <summary>
    /// Format PICC
    /// Need Authentication for PICC / Application 0x000000
    /// </summary>
    fn format_picc(&self) -> Result<()> {
        let cmd_format = APDUCommand {
            case: IsoCase::Case2Short,
            cla: 0x90,
            ins: APDUInstructions::FORMAT_PICC as u8,
            ..Default::default()
        };

        let response = self.card.transmit(cmd_format).unwrap();

        response.check()
    }

    /// <summary>
    /// Create Application for ApplicationID
    /// </summary>
    /// <param name="aid">3 Byte ID</param>
    fn create_application(&self, aid: u32, keysetting1: u8, keysetting2: u8) -> Result<()> {
        if aid > 0xFFFFFF {
            return Err(InvalidApplicationID);
        }

        let mut data = aid.to_le_bytes()[..3].to_vec();
        data.push(keysetting1);
        data.push(keysetting2);
        let cmd_create_application = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::CREATE_APPLICATION as u8,
            data: Option::from(data), //FIXME: Which byteorder?
            ..Default::default()
        };

        let response = self.card.transmit(cmd_create_application).unwrap();

        response.check()
    }

    /// <summary>
    /// Change AES key, the same as Authenticated
    /// </summary>
    /// <param name="key_id">0x01 - 0x0D</param>
    /// <param name="new_key">Array of 16 Bytes</param>
    /// <param name="key_version">Version of Key(min. 0x10)</param>
    fn change_key_aes(&mut self, key_id: u8, new_key: &[u8], key_version: u8) -> Result<()> {
        if key_id >= 0x0E {
            return Err(InvalidKeyID);
        }

        let header = vec![0xC4, key_id];
        let key_and_version: Vec<u8> = [new_key, &[key_version]].concat();
        let mut command = vec![];

        command.extend(&header);
        command.extend(&key_and_version);

        let crc = crate::crypto::crc::crc32::calculate(command.as_slice());

        let mut plaintext: Vec<u8> = vec![];
        plaintext.extend(key_and_version);
        plaintext.extend(crc);

        let plaintext_pad = expand_to_blocksize(plaintext.as_mut_slice(), 16)?;

        let cryptogram = AES::encrypt(plaintext_pad, self.session_key.as_ref().unwrap(), self.cbc_iv.as_ref().unwrap())?;

        self.cbc_iv = Some(util::extract_last_block(cryptogram.as_slice(), 16)?.to_vec());

        let mut data: Vec<u8> = vec![key_id];
        data.extend(cryptogram);

        let cmd_change_key = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::CHANGE_KEY as u8,
            data: Option::from(data), //FIXME: Which byteorder?
            ..Default::default()
        };

        let response = self.card.transmit(cmd_change_key).unwrap();

        response.check()
    }

    /// <summary>
    /// Change AES key, other than Authenticated
    /// </summary>
    /// <param name="key_id">0x01 - 0x0D</param>
    /// <param name="new_key">Array of 16 Bytes</param>
    /// <param name="old_key">Array of 16 Bytes</param>
    /// <param name="key_version">Version of Key(min. 0x10)</param>
    fn change_other_key_aes(&mut self, key_id: u8, new_key: &[u8], old_key: &[u8], key_version: u8) -> Result<()> {
        if key_id >= 0x0E {
            return Err(InvalidKeyID);
        }

        let header = vec![0xC4, key_id];

        let key_xor: Vec<u8> = new_key.iter().zip(old_key.iter()).map(|(&x1, &x2)| x1 ^ x2).collect();

        let key_and_version: Vec<u8> = [key_xor, vec![key_version]].concat();
        let mut command = vec![];

        command.extend(&header);
        command.extend(&key_and_version);

        let crc = crate::crypto::crc::crc32::calculate(command.as_slice());

        let mut plaintext: Vec<u8> = vec![];
        plaintext.extend(key_and_version);
        plaintext.extend(crc);

        let plaintext_pad = expand_to_blocksize(plaintext.as_mut_slice(), 16)?;

        let cryptogram = AES::encrypt(plaintext_pad, self.session_key.as_ref().unwrap(), self.cbc_iv.as_ref().unwrap())?;

        self.cbc_iv = Some(util::extract_last_block(cryptogram.as_slice(), 16)?.to_vec());

        let mut data: Vec<u8> = vec![key_id];
        data.extend(cryptogram);

        let cmd_change_key = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::CHANGE_KEY as u8,
            data: Option::from(data), //FIXME: Which byteorder?
            ..Default::default()
        };

        let response = self.card.transmit(cmd_change_key).unwrap();

        response.check()
    }

    fn create_file_standard(&self, file_id: u8, communication: FileCommunication, access_rights: u16, size: u32) -> Result<()> {
        if file_id >= 0x20 {
            return Err(InvalidFileID);
        }

        let access_rights = access_rights.to_le_bytes().to_vec();
        let size = size.to_le_bytes()[..3].to_vec();
        let mut data = vec![file_id, communication as u8];
        data.extend(&access_rights);
        data.extend(&size);

        let cmd_create_file_standard = APDUCommand {
            case: IsoCase::Case4Short,
            cla: 0x90,
            ins: APDUInstructions::CREATE_STDDATAFILE as u8,
            data: Option::from(data),
            ..Default::default()
        };

        let response = self.card.transmit(cmd_create_file_standard).unwrap();

        response.check()
    }

    /// <summary>
    /// Write Data to File
    /// </summary>
    /// <param name="file_id">ID of File (0x00 - 0x20)</param>
    /// <param name="offset">Offset for File</param>
    /// <param name="data">Data to write</param>
    fn write_data(&self, file_id: u8, offset: u32, data: &[u8]) -> Result<()> {
        if file_id >= 0x20 {
            return Err(InvalidFileID);
        }

        const MAX_BYTES_PER_TRANSACTION: usize = 47;
        let mut bytes_writen: usize = 0;
        let length: usize = data.len();

        let mut ret: Result<()> = Err(error::Error::Simple(simple_error::simple_error!("Error while writing data to card"))); //TODO: Replace with better error

        while bytes_writen != data.len() {
            let bytes_towrite = match length - bytes_writen < MAX_BYTES_PER_TRANSACTION {
                true => { length - bytes_writen }
                false => { MAX_BYTES_PER_TRANSACTION }
            };

            let mut write_buffer = vec![file_id];
            write_buffer.append(&mut offset.to_le_bytes()[..3].to_vec());
            write_buffer.append(&mut bytes_towrite.to_le_bytes()[..3].to_vec());
            write_buffer.append(&mut data[bytes_writen..bytes_writen + bytes_towrite].to_vec());
            bytes_writen += bytes_towrite;

            let cmd_write_data = APDUCommand {
                case: IsoCase::Case4Short,
                cla: 0x90,
                ins: APDUInstructions::WRITE_DATA as u8,
                data: Option::from(write_buffer),
                ..Default::default()
            };

            let response = self.card.transmit(cmd_write_data).unwrap();

            ret = response.check();
        };

        return ret;
    }

    /// <summary>
    /// Read Data from File
    /// </summary>
    /// <param name="file_id">ID of File (0x00 - 0x20)</param>
    /// <param name="offset">Offset for File</param>
    /// <param name="length">Lenght of Data</param>
    fn read_data(&self, file_id: u8, offset: u32, length: usize) -> Result<Vec<u8>> {
        if file_id >= 0x20 {
            return Err(InvalidFileID);
        }

        const MAX_BYTES_PER_TRANSACTION: usize = 47;
        let mut bytes_read: usize = 0;

        let mut read_buffer: Vec<u8> = vec![];

        while bytes_read != length {
            let bytes_toread = match length - bytes_read < MAX_BYTES_PER_TRANSACTION {
                true => { length - bytes_read }
                false => { MAX_BYTES_PER_TRANSACTION }
            };

            let mut send_buffer = vec![file_id];
            send_buffer.append(&mut offset.to_le_bytes()[..3].to_vec());
            send_buffer.append(&mut bytes_toread.to_le_bytes()[..3].to_vec());
            bytes_read += bytes_toread;

            let cmd_read_data = APDUCommand {
                case: IsoCase::Case4Short,
                cla: 0x90,
                ins: APDUInstructions::READ_DATA as u8,
                data: Option::from(send_buffer),
                ..Default::default()
            };

            let response = self.card.transmit(cmd_read_data).unwrap();

            response.check().or_else(|e| return Err(e));

            read_buffer.append(&mut response.body.unwrap()[..bytes_toread].to_vec());
        };

        return Ok(read_buffer);

    }
}


/// <summary>
/// Generates SessionKey for DES Authentification
/// </summary>
/// <returns>8Byte SessionKey</returns>
fn generate_session_key_des(rnd_a: &[u8], rnd_b: &[u8]) -> Option<Vec<u8>> {
    let sessionkey = [&rnd_a[..4], &rnd_b[..4]].concat();
    Some([sessionkey.as_slice(), sessionkey.as_slice()].concat())
}

/// <summary>
/// Generates SessionKey for AES Authentification
/// </summary>
/// <returns>16Byte SessionKey</returns>
fn generate_session_key_aes(rnd_a: &[u8], rnd_b: &[u8]) -> Option<Vec<u8>> {
    //FIXME: Check math
    Some([&rnd_a[..4], &rnd_b[..4], &rnd_a[12..], &rnd_b[12..]].concat())
}

/// <summary>
/// Generate KeySetting1 for Application Settings or PICC Setting
/// </summary>
/// <param name="changeKey">ID of Key for changing Application Keys</param>
/// <returns>generated keysettings</returns>
fn generate_keysetting1(change_key: u8, change_masterkey_settings: ChangeMasterKeySettings, create_delete_file: CreateDeleteFile, file_directory_access: FileDirectoryAccess, change_master_key: ChangeMasterKey) -> Result<u8> {
    return match FromPrimitive::from_u8(change_key) {
        Some(ChangeApplicationKey::MASTERKEY) | Some(ChangeApplicationKey::SAMEKEY) | Some(ChangeApplicationKey::ALLKEYS) => Ok((change_key << 4) | change_masterkey_settings as u8 | create_delete_file as u8 | file_directory_access as u8 | change_master_key as u8),
        None => {
            if change_key < 0x01 || change_key >= 0x08 {
                Err(InvalidKeyID)
            } else {
                Ok((change_key << 4) | change_masterkey_settings as u8 | create_delete_file as u8 | file_directory_access as u8 | change_master_key as u8)
            }
        }
    };
}

/// <summary>
/// Generate KeySetting2 for Application Creation
/// </summary>
/// <param name="numberOfKeys">Number of keys that can be stored within the application (0x01-0x0D)</param>
/// <returns>generated keysettings</returns>
fn generate_keysetting2(crypto_operations: CryptoOperationsType, file_identifier: FileIdentifiers, num_keys: u8) -> Result<u8> {
    return if num_keys < 0x01 || num_keys >= 0x0D {
        Err(NumKeys)
    } else {
        Ok(crypto_operations as u8 | file_identifier as u8 | num_keys)
    };
}

/// <summary>
/// Generate FileAccess Rights for File Settings
/// Use enum AccesRights for Free or Never Option
/// </summary>
/// <param name="read">KeyID for Read Access</param>
/// <param name="write">KeyID for Write Access</param>
/// <param name="read_write">KeyID for Read and Write Access</param>
/// <param name="configure">KeyID for Configuration Access</param>
fn generate_file_access_rights(read: u8, write: u8, read_write: u8, configure: u8) -> Result<u16> {
    return if read > 0x0F || write >= 0x0F || read_write >= 0x0F || configure >= 0x0F {
        Err(InvalidKeyID)
    } else {
        Ok((((read as u16) << 12) | ((write as u16) << 8) | ((read_write as u16) << 4) | configure as u16) as u16)
    };
}


#[cfg(test)]
mod tests {
    use hex_literal::hex;
    use mockall::{mock, predicate};
    use crate::{APDUResponse, APDUCommand, Card};
    use crate::error::Result;
    use crate::desfire::desfire::{generate_session_key_des, generate_session_key_aes, generate_keysetting1, generate_keysetting2, generate_file_access_rights, Desfire};
    use crate::desfire::{ChangeApplicationKey, ChangeMasterKeySettings, CreateDeleteFile, FileDirectoryAccess, ChangeMasterKey, CryptoOperationsType, FileIdentifiers};
    use crate::error::Error::InvalidKeyID;

    mock! {
        pub VirtualCard {}
        impl Card for VirtualCard {
            fn connect(&self);
            fn disconnect(&self);
            fn transmit(&self, apdu_cmd: APDUCommand) -> Result<APDUResponse>;
        }
    }

    #[test]
    fn generate_sessionkey_des() {
        let rndA = hex!("a541a9dc9138df07");
        let rndB = hex!("cbe55aa893b2da25");

        let expected_sessionkey = hex!("a541a9dccbe55aa8a541a9dccbe55aa8");

        let sessionkey = generate_session_key_des(&rndA, &rndB).unwrap();

        println!("expected sessionkey: {:X?}", expected_sessionkey);
        println!("actual sessionkey:   {:X?}", sessionkey.as_slice());

        assert_eq!(expected_sessionkey, sessionkey.as_slice());
    }

    #[test]
    fn generate_sessionkey_aes() {
        let rndA = hex!("bc14dfde20074617e45a8822f06fdd91");
        let rndB = hex!("482ddc54426e6dee560413b8d95471f5");

        let expected_sessionkey = hex!("bc14dfde482ddc54f06fdd91d95471f5");

        let sessionkey = generate_session_key_aes(&rndA, &rndB).unwrap();

        println!("expected sessionkey: {:X?}", expected_sessionkey);
        println!("actual sessionkey:   {:X?}", sessionkey.as_slice());

        assert_eq!(expected_sessionkey, sessionkey.as_slice());
    }

    #[test]
    fn generate_key_setting1() {
        assert_eq!(0x0B,
                   generate_keysetting1(ChangeApplicationKey::MASTERKEY as u8,
                                        ChangeMasterKeySettings::WITHMASTERKEY,
                                        CreateDeleteFile::ONLYMASTERKEY,
                                        FileDirectoryAccess::NOKEY,
                                        ChangeMasterKey::CHANGEABLE).unwrap()
        )
    }

    #[test]
    fn generate_key_setting1_changekey() {
        assert_eq!(0x1B,
                   generate_keysetting1(0x01,
                                        ChangeMasterKeySettings::WITHMASTERKEY,
                                        CreateDeleteFile::ONLYMASTERKEY,
                                        FileDirectoryAccess::NOKEY,
                                        ChangeMasterKey::CHANGEABLE).unwrap()
        )
    }

    #[test]
    #[should_panic(expected= "InvalidKeyID")]
    fn generate_key_setting1_invalid_keyid() {
                   generate_keysetting1(0x10,
                                        ChangeMasterKeySettings::WITHMASTERKEY,
                                        CreateDeleteFile::ONLYMASTERKEY,
                                        FileDirectoryAccess::NOKEY,
                                        ChangeMasterKey::CHANGEABLE).unwrap();
    }

    #[test]
    fn generate_key_setting2() {
        assert_eq!(0x82, generate_keysetting2(CryptoOperationsType::AES, FileIdentifiers::NOTUSED, 0x02).unwrap());
    }

    #[test]
    #[should_panic(expected= "NumKeys")]
    fn generate_key_setting2_wrong_num_keys() {
        generate_keysetting2(CryptoOperationsType::AES, FileIdentifiers::NOTUSED, 0x10).unwrap();
    }

    #[test]
    fn generate_file_accessrights() {
        assert_eq!(0x1234, generate_file_access_rights(0x01, 0x02, 0x03, 0x04).unwrap())
    }

    #[test]
    #[should_panic(expected = "InvalidKeyID")]
    fn generate_file_accessrights_outofrange() {
        generate_file_access_rights(0x10, 0x00, 0x00, 0x00).unwrap();
    }

    #[test]
    fn select_application() {
        let mut mock = MockVirtualCard::new();
        mock.expect_transmit().withf(|x: APDUCommand| (x as Vec<u8>) == vec![0x90, 0x5a, 0x00, 0x00, 0x03, 0x33, 0x22, 0x11, 0x00]).return_const(Ok(APDUResponse{
            body: None,
            sw1: 0x91,
            sw2: 0x00
        }));

        let desfire = Desfire{
            card: Box::new(mock),
            cbc_iv: None,
            session_key: None
        };

        assert!(desfire.select_application(0x112233).is_ok())


    }
}